"""
Original © NetApp 2024

This package contains useful output formatters. An output formatter can be defined
as the return type of a CLI command (using a type annotation). If one is defined,
then the result of running the command will be passed to the output formatter
which will usually print the result to the screen for the user.
"""
